library("ggplot2")
library("reshape2")
library("plyr")
library("scales")

read_cpu_data = function(filename_prefix, filename_suffix) {
  for (i in 1:5) {
    filename = paste(filename_prefix, as.character(i), filename_suffix, sep="");
    df.one = read.table(filename, header = T)
    df.one$run = i
    if (i == 1) {
      df = df.one;
    } else {
      df <- rbind(df, df.one)
    }
  }
  
  cn = colnames(df)
  time.names = cn[startsWith(cn, "time.")]
  df = melt(df, c("bench", "run"), time.names, value.name = "time", variable.name = "schedule")
  df = merge(df[df$schedule == "time.orig",], df[df$schedule != "time.orig",], by=c("bench", "run")) 
  df$speedup = df$time.x / df$time.y
  
  df = subset(df, select = c("bench", "run", "schedule.y", "speedup"))
  colnames(df)[3] = "schedule"
  df
}

read_gpu_data = function(filename_prefix, filename_suffix) {
  for (i in 1:4) {
    filename = paste(filename_prefix, as.character(i), filename_suffix, sep="");
    df.one = read.table(filename, col.names = c("bench", "time.kernel.public",  "launches.public",  "kernels.public",  "time.all.public",
                                                         "time.kernel.trunk",   "launches.trunk",   "kernels.trunk",   "time.all.trunk",
                                                         "time.kernel.spatial", "launches.spatial", "kernels.spatial", "time.all.spatial"))
    df.one$run = i
    if (i == 1) {
      df = df.one;
    } else {
      df = rbind(df, df.one)
    }
  }
  
  cn = colnames(df)
  time.kernel.names = cn[startsWith(cn, "time.kernel")]
  time.all.names = cn[startsWith(cn, "time.all")]
  launches.names = cn[startsWith(cn, "launches.")]
  kernels.names = cn[startsWith(cn, "kernels.")]
  
  df.times.kernel = subset(df, select = c("bench", "run", time.kernel.names))
  df.times.all = subset(df, select = c("bench", "run", time.all.names))
  df.launches = subset(df, select = c("bench", "run", launches.names))
  df.kernels = subset(df, select = c("bench", "run", kernels.names))
  
  df.times.kernel = melt(df.times.kernel, c("bench", "run"), time.kernel.names, value.name ="time.kernel", variable.name = "schedule")
  df.times.kernel$schedule = as.factor(gsub("time.kernel.", "", as.character(df.times.kernel$schedule)))
  df.times.all = melt(df.times.all, c("bench", "run"), time.all.names, value.name ="time.all", variable.name = "schedule")
  df.times.all$schedule = as.factor(gsub("time.all.", "", as.character(df.times.all$schedule)))
  df.launches = melt(df.launches, c("bench", "run"), launches.names, value.name ="launches", variable.name = "schedule")
  df.launches$schedule = as.factor(gsub("launches.", "", as.character(df.launches$schedule)))
  df.kernels = melt(df.kernels, c("bench", "run"), kernels.names, value.name ="kernels", variable.name = "schedule")
  df.kernels$schedule = as.factor(gsub("kernels.", "", as.character(df.kernels$schedule)))
  
  df = merge(df.times.kernel, df.times.all, by=c("bench", "run", "schedule"))
  df = merge(df, df.launches, by=c("bench", "run", "schedule"))
  df = merge(df, df.kernels, by=c("bench", "run", "schedule"))
}

log10p = function(x) { log(1 + x, 10) }
exp10m = function(x) { 10**x - 1 }
log10p_trans = trans_new("log10p", log10p, exp10m, domain=c(-1,Inf))

mediancilow = function(x){ z = apply(matrix(sample(x, rep=TRUE, 10^4*length(x)), nrow=10^4), 1, median); quantile(z, 0.025, na.rm=T) }
mediancihigh = function(x){ z = apply(matrix(sample(x, rep=TRUE, 10^4*length(x)), nrow=10^4), 1, median); quantile(z, 0.975, na.rm=T) }

meancilow = function(x){ z = apply(matrix(sample(x, rep=TRUE, 10^4*length(x)), nrow=10^4), 1, mean); quantile(z, 0.025, na.rm=T) }
meancihigh = function(x){ z = apply(matrix(sample(x, rep=TRUE, 10^4*length(x)), nrow=10^4), 1, mean); quantile(z, 0.975, na.rm=T) }

popl_theme =  theme_bw() +
  theme(axis.text.x = element_text(angle=45, vjust = 1, hjust = 1),
        legend.position = c(0.99,0.99),
        legend.justification = c("right", "top"),
        legend.direction = "horizontal",
        legend.key.size = unit(6, units="pt"),
        legend.title = element_text(size=10),
        legend.box.background = element_rect(linetype = 1))

dd = read_cpu_data("cpu_ficus_", ".txt")
dd$schedule = as.factor(as.character(dd$schedule))
dd$schedule = factor(dd$schedule, levels(dd$schedule)[c(1,3,5,4,2)])
dd$schedule = revalue(dd$schedule, c("time.pluto"="Pluto",
                                     "time.trunk"="ppcg trunk",
                                     "time.public"="ppcg public",
                                     "time.spatial"="ppcg spatial",
                                     "time.posttile"="ppcg spatial + post-tile"))

# dda <- ddply(dd, ~ schedule + bench, summarize, median = mean(speedup), ci.low = meancilow(speedup), ci.high = meancihigh(speedup))

dda <- ddply(dd, ~schedule+bench, summarize, speedup=median(speedup))
ggplot(dda, log10="y") +
  geom_bar(aes(x=bench, y=speedup, fill=schedule, group=schedule), position=position_dodge(0.8), stat="identity", color="gray70") +
#  geom_errorbar(aes(ymin=ci.low, ymax=ci.high, x=bench, group=schedule), position=position_dodge(0.8)) +
  geom_hline(yintercept = 1, color="gray50") +
  xlab("Benchmark") +
  scale_y_continuous(name = "Speedup over original (log scale)", trans = log10p_trans) +
  popl_theme +
  scale_fill_brewer(palette='OrRd')

ggsave("../img/ficus_cpu.pdf", units="cm", width=24, height=8)

dd = read_cpu_data("cpu_serial_aleph_",".txt")
dd$schedule = as.factor(as.character(dd$schedule))
dd$schedule = revalue(dd$schedule, c("time.serial.pluto"="Pluto",
                                     "time.serial.public"="ppcg public",
                                     "time.serial.trunk"="ppcg trunk",
                                     "time.serial2"="ppcg spatial",
                                     "time.serial2.posttile"="ppcg spatial + post-tile"))
dda <- ddply(dd, ~schedule+bench, summarize, speedup=median(speedup))
ggplot(dda) + 
  geom_bar(aes(x=bench, y=speedup, fill=schedule, group=schedule), position = position_dodge(0.8), stat="identity", color="gray70") +
  geom_hline(yintercept = 1, color="gray50") +
  xlab("Benchmark") +
  scale_y_continuous(name = "Speedup over original (log scale)", trans = log10p_trans, limits = c(0,7)) +
  popl_theme +
  scale_fill_brewer(palette='OrRd')
ggsave("../img/aleph_cpu.pdf", units="cm", width=24, height=8)

#wilcox.test(speedup ~ schedule, data=dd, alternative="less", subset=schedule %in% c("Pluto", "ppcg spatial"))

dd <- read_cpu_data("cpu_jun_", ".txt")
dd = dd[dd$schedule != "time.pluto",]
dd$schedule = as.factor(as.character(dd$schedule))
dd$schedule = factor(dd$schedule, levels(dd$schedule)[c(1,2,4,6,5,3)])
dd$schedule = revalue(dd$schedule, c("time.pluto.noprevec"="Pluto",
                                     "time.polyast"="PolyAST w/o red",
                                     "time.public"="ppcg public",
                                     "time.trunk"="ppcg trunk",
                                     "time.spatial"="ppcg spatial",
                                     "time.posttile"="ppcg spatial + post-tile"))

dda <- ddply(dd, ~schedule+bench, summarize, speedup=median(speedup))
ggplot(dda) + 
  geom_bar(aes(x=bench, y=speedup, fill=schedule, group=schedule), position = position_dodge(0.8), stat="identity", color="gray70") +
  geom_hline(yintercept = 1, color="gray50") +
  xlab("Benchmark") +
  scale_y_continuous(name = "Speedup over original (log scale)", trans = log10p_trans, limits = c(0,32)) +
  popl_theme +
  scale_fill_brewer(palette='OrRd')
ggsave("../img/jun_cpu.pdf", units="cm", width=24, height=8)
